/*
 * PIMPPA - adopter
 *
 * Adopter goes through matching fileareas (default=all) and adds 
 * missing files to the databases. (==File exists in directory, 
 * but not in database)
 *
 *
 */

#include "pimppa.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <ctype.h>

int Verbose=0;

char AreaPattern[256]="%";				// Areas to operate on. Def=all

struct fileareas
{
	char area_filepath[PATH_MAX];
	int area_id;
	struct fileareas *next;
};

void usage(char *name)
{
	fprintf(stderr, "Usage: %s <options>\n\n"
					"-a <pattern>     Area SQL pattern to operate on, def=all\n"
					"-m               Move from current dir\n"
					"-v               Verbose execution\n", name);
}

int move(void);

int adopt(void)
{
	MYSQL db;
	MYSQL_RES *sql_res;
	MYSQL_ROW sql_row;
	int added_files=0, added_per_area=0;
	struct fileareas *areas=NULL;
	struct fileareas *tmp;
	DIR *dirri;
	struct dirent *dirdata;
	struct stat st;
	unsigned long file_id=0;

	if(!p_connect(&db))
		return(0);

/******** find directories of matching fileareas  *******/

	fprintf(stderr, "Browsing fileareas...\n");

	p_query(&db, "SELECT area_path, area_id 
					FROM p_areas 
					WHERE area_name like '%s'
					ORDER BY area_name DESC",
					AreaPattern);
	if(mysql_error(&db)[0])
		return(-1);
	sql_res=mysql_use_result(&db);

	while((sql_row=mysql_fetch_row(sql_res)))
	{
		tmp=malloc(sizeof(struct fileareas));
		strcpy(tmp->area_filepath, sql_row[0]);
		if(!p_checkp(tmp->area_filepath))
			strcat(tmp->area_filepath, "/");
		tmp->area_id=atoi(sql_row[1]);
		tmp->next=NULL;

		if(!areas)
			areas=tmp;
		else
		{
			tmp->next=areas;
			areas=tmp;
		}
	}
	mysql_free_result(sql_res);
	
/**** go through the areas ******/

	tmp=areas;
	while(tmp)
	{
		added_per_area=0;

		fprintf(stderr, "Checking %s ", tmp->area_filepath);
		
		chdir(tmp->area_filepath);

		dirri=opendir(tmp->area_filepath);
		if(!dirri)
		{	
			fprintf(stderr, "Unable to open dir %s\n", tmp->area_filepath);
			tmp=tmp->next;
			free(areas);
			areas=tmp;
			continue;
		}

		while((dirdata=readdir(dirri)))
		{
			char pattern[P_LEN_FILE_NAME];
			char escaped_fn[2*P_LEN_FILE_NAME+1];

			if(stat(dirdata->d_name, &st)!=0)
			{
				fprintf(stderr, "stat() error on %s%s\n", 
						tmp->area_filepath, dirdata->d_name);
				continue;
			}
			
			if(!(S_ISREG(st.st_mode) || S_ISLNK(st.st_mode))
				|| S_ISDIR(st.st_mode))
				continue;

			file_id=0;

			// Duplicate check
			mysql_escape_string(escaped_fn, dirdata->d_name, strlen(dirdata->d_name));

			p_query(&db, "SELECT file_integ, UNIX_TIMESTAMP(file_date),
							file_id FROM p_files
							WHERE file_name='%s' AND file_area=%d",
							escaped_fn, tmp->area_id);
			if(mysql_error(&db)[0])
			{
				fprintf(stderr, "%s\n", mysql_error(&db));
				continue;
			}
			sql_res=mysql_store_result(&db);
			if(sql_res && (sql_row=mysql_fetch_row(sql_res)))// File exists?
			{
				if(atoi(sql_row[0])==INTEG_FAILED)
				{
					if(atol(sql_row[1])>=st.st_mtime)	// File seems the same
					{
						mysql_free_result(sql_res);
						continue;
					}
					file_id=atol(sql_row[2]);
				}
				else
				{
					mysql_free_result(sql_res);
					continue;
				}
			}
			mysql_free_result(sql_res);

			if(Verbose)
				printf("On %d found %s missing, adding\n", 
						tmp->area_id, dirdata->d_name);
	
			if(!file_id)
			{
				p_query(&db, "INSERT INTO p_files
								(file_name, file_area, file_size, file_date) 
								VALUES ('%s',%d,%ld,FROM_UNIXTIME(%ld))",
								escaped_fn, tmp->area_id, st.st_size,
								st.st_mtime);
			}
			else
			{
				p_query(&db, "REPLACE INTO p_files
								(file_id, file_name, file_area, file_size, 
								file_date, file_backup) 
								VALUES (%ld,'%s',%d,%ld,FROM_UNIXTIME(%ld),0)",
								file_id, escaped_fn, tmp->area_id,
								st.st_size, st.st_mtime);
			}
			if(mysql_error(&db)[0])
				fprintf(stderr, "Unable to add %s to %d (%s)\n", 
						dirdata->d_name, tmp->area_id, mysql_error(&db));
			else
			{
				added_files++;
				added_per_area++;
			}

			if(p_makepattern(pattern, dirdata->d_name)>=P_MIN_ASS_NAMELENGTH)
			{
				char escaped_pat[2*P_LEN_FILE_NAME+1];

				mysql_escape_string(escaped_pat, pattern, strlen(pattern));

				p_query(&db, "REPLACE INTO p_assign (a_pattern, a_dest)
								VALUES ('%s', %d)", escaped_pat,
								tmp->area_id);
				if(mysql_error(&db)[0])
					fprintf(stderr, "%s\n", mysql_error(&db));
			}

		}
		closedir(dirri);
		
		tmp=tmp->next;
		free(areas);
		areas=tmp;

		if(added_per_area)
			fprintf(stderr, "(%d added)\n", added_per_area);
		else
			fprintf(stderr, "\n");
	}

	mysql_close(&db);
	return(added_files);
}
	

int main(int argc, char *argv[])
{
	int added_files;
	int domove=0;
	int go=1;
	char olddir[PATH_MAX];
	
	while(go)
	{
		switch(getopt(argc, argv, "a:hmvV"))
		{
			case 'a':
				strcpy(AreaPattern, optarg);
				break;
			case 'h':
				usage(argv[0]);
				return(0);
				break;
			case 'm':
				domove=1;
				break;
			case 'v':
				Verbose=1;
				break;
			case 'V':
				printf("%s %s %s\n", PACKAGE, argv[0], VERSION);
				return(0);
				break;
			default:
				go=0;
				break;
		}
	}	

	getcwd(olddir, PATH_MAX);

	if(!domove)
	{
		added_files=adopt();
		printf("Done, %d files added.\n", added_files);
	}
	else
		move();

	chdir(olddir);

	return(0);
}

/*
 * This moves files from the current directory to appropriate
 * fileareas, if assign pattern is known for them.
 *
 */
int move(void)
{
	MYSQL db;
    MYSQL_RES *sql_res;
    MYSQL_ROW sql_row;
    DIR *dirp;
    struct dirent *dirdata;
    int total_files=0,total_bytes=0;
    int assigned_files=0,assigned_bytes=0;
	
	if(!p_connect(&db))
		return(0);

    dirp=opendir(".");
    if(!dirp)
        return(0);

    while((dirdata=readdir(dirp)))
    {
        char *filename=dirdata->d_name;
        struct stat st;
        char pattern[P_LEN_FILE_NAME];
        char escaped_fn[2*P_LEN_FILE_NAME+1];
        int destarea_id, moveok;
		
        destarea_id=moveok=0;

        stat(filename, &st);
        if(!(S_ISREG(st.st_mode) || S_ISLNK(st.st_mode))
            || S_ISDIR(st.st_mode))
            continue;
			
		total_files++;
        total_bytes+=st.st_size;
        
		if(Verbose)
            printf("%s ", filename);

		// Find out destination area
        if(p_makepattern(pattern, filename)>=P_MIN_ASS_NAMELENGTH)
        {
            char escaped_pat[2*P_LEN_FILE_NAME+1];

            mysql_escape_string(escaped_pat, pattern, strlen(pattern));

            // Locate the destination area id
            p_query(&db, "SELECT a_dest FROM p_assign
                          WHERE a_pattern='%s'", escaped_pat);
            sql_res=mysql_store_result(&db);
            if(sql_res)
            {
                if((sql_row=mysql_fetch_row(sql_res)))
                    destarea_id=atoi(sql_row[0]);
                mysql_free_result(sql_res);
            }
        }

        if(destarea_id<=0) // Unknown file, do nothing
        {
            if(Verbose)
				printf("- Destination not known (%d)\n", destarea_id);
        	continue;
		}
		
		// Duplicate check
		mysql_escape_string(escaped_fn, filename, strlen(filename));
		p_query(&db, "SELECT file_integ FROM p_files
						WHERE file_name='%s' AND file_area=%d",
						escaped_fn, destarea_id);
		if(mysql_error(&db)[0])
		{
			fprintf(stderr, "%s\n", mysql_error(&db));
			continue;
		}
		sql_res=mysql_store_result(&db);
		if(sql_res && (sql_row=mysql_fetch_row(sql_res)))	// File exists?
		{
			if(atoi(sql_row[0])!=INTEG_FAILED)
			{
				if(Verbose)
					printf(" - Already exists on %d!\n", destarea_id);
				
				mysql_free_result(sql_res);
				continue;
			}
			else
			{
				if(Verbose)
					printf(" Exists on %d, but broken, replacing. ",
							destarea_id);
			}
		}
		if(sql_res)
			mysql_free_result(sql_res);

		// Generate old & new pathnames, physical move
        p_query(&db, "SELECT area_path FROM p_areas
                      WHERE area_id=%d",
                      destarea_id);
        sql_res=mysql_store_result(&db);
        if(sql_res)
        {
            if((sql_row=mysql_fetch_row(sql_res)))  // Ok, assign!
            {
                char newpath[PATH_MAX];
				char oldpath[PATH_MAX];

                sprintf(newpath, "%s%s%s", sql_row[0], 
						(p_checkp(sql_row[0]) ? "" : "/"),
						filename);
				sprintf(oldpath, "./%s", filename);
		
				if(!p_mv(oldpath, newpath))
				{
					fprintf(stdout, "p_mv() error from %s to %s\n",
							filename, newpath);
					
					mysql_free_result(sql_res);
					continue;
				}
				moveok=1;
			}
			mysql_free_result(sql_res);
		}
		if(!moveok)
		{
			fprintf(stdout, "Destination area %d not found...\n",
					destarea_id);
			continue;
		}

 		assigned_files++;
		assigned_bytes+=st.st_size;

		if(Verbose)
			printf("- Assign to %d ", destarea_id);
		
		if(destarea_id>0 && moveok)
		{
			p_query(&db, "REPLACE INTO p_files (file_name, file_size,
        		           file_area,file_date,file_backup)
                           VALUES ('%s',%ld,%d,FROM_UNIXTIME(%ld),0)",
                           escaped_fn, st.st_size, destarea_id,
                           (long) time(NULL));
//			printf("\n%s\n", query);
		}

        if(Verbose)
            printf("- Ok\n");
	}

    printf("Done, %d/%d files and %d/%d bytes assigned to fileareas.\n"
            "%d files and %d bytes were skipped.\n",
            assigned_files,total_files, assigned_bytes,total_bytes,
            total_files-assigned_files, total_bytes-assigned_bytes);

	mysql_close(&db);

    return(total_files-assigned_files);
}

